import logging
from datetime import datetime, timedelta
from typing import Optional

import numpy as np
import pandas as pd
from statsmodels.api import OLS

from timetomodel import modelling, speccing, transforming

logger = logging.getLogger(__name__)


def create_dummy_model_state(
    data_start: datetime,
    data_range_in_hours: int,
    outcome_feature_transformation: Optional[
        transforming.ReversibleTransformation
    ] = None,
    regressor_feature_transformation: Optional[
        transforming.ReversibleTransformation
    ] = None,
) -> modelling.ModelState:
    """
    Create a dummy model. data increases linearly, regressor is constant (useless).
    Use two different ways to define Series specs to test them.
    """
    dt_range = pd.date_range(
        data_start,
        data_start + timedelta(hours=data_range_in_hours),
        closed="left",
        freq="1H",
    )
    reg_range = pd.date_range(
        data_start,
        data_start + timedelta(hours=data_range_in_hours) + timedelta(days=1),
        closed="left",
        freq="1H",
    )  # 1 additional day of regressor data is available
    outcome_values = [0]
    regressor_values = [5]
    for i in range(1, len(dt_range)):
        outcome_values.append(outcome_values[i - 1] + 1)
    for i in range(1, len(reg_range)):
        regressor_values.append(5)
    outcome_series = pd.Series(index=dt_range, data=outcome_values)
    regressor_series = pd.Series(index=reg_range, data=regressor_values)
    specs = modelling.ModelSpecs(
        outcome_var=speccing.ObjectSeriesSpecs(
            outcome_series,
            name="my_outcome",
            feature_transformation=outcome_feature_transformation,
        ),
        model=OLS,
        lags=[2, 3],  # lags of interest are 2 and 3 hours (we'll ignore the last hour)
        frequency=timedelta(hours=1),
        horizon=timedelta(minutes=240),
        remodel_frequency=timedelta(hours=48),
        regressors=[
            speccing.ObjectSeriesSpecs(
                regressor_series,
                "my_regressor",
                feature_transformation=regressor_feature_transformation,
            )
        ],
        start_of_training=data_start
        + timedelta(hours=3),  # leaving room for NaN in lags
        end_of_testing=data_start + timedelta(hours=int(data_range_in_hours / 3)),
    )
    return modelling.ModelState(
        modelling.create_fitted_model(specs, version="0.1"), specs
    )


class MyDFPostProcessing(transforming.Transformation):
    def transform_dataframe(self, df: pd.DataFrame):
        """Keep the most recent observation, drop duplicates"""
        return (
            df.sort_values(by=["horizon"], ascending=True)
            .drop_duplicates(subset=["datetime"], keep="first")
            .sort_values(by=["datetime"])
        )


class MyAdditionTransformation(transforming.ReversibleTransformation):
    def transform_series(self, x: pd.Series):
        logger.debug("Adding %s to %s ..." % (self.params.addition, x))
        return x + self.params.addition

    def back_transform_value(self, y: np.array):
        logger.debug("Subtracting %s from %s ..." % (self.params.addition, y))
        return y - self.params.addition


class MyMultiplicationTransformation(transforming.ReversibleTransformation):
    def transform_series(self, x: pd.Series):
        return x * self.params.factor

    def back_transform_value(self, y: np.array):
        return y / self.params.factor


class MyNanIntroducingTransformation(transforming.ReversibleTransformation):
    def transform_series(self, x: pd.Series) -> pd.Series:
        x.iloc[5] = np.nan
        return x

    def back_transform_value(self, y: np.array) -> np.array:
        return y



import logging
from datetime import datetime, timedelta

import pandas as pd
import pytest
import pytz
from statsmodels.tools.sm_exceptions import MissingDataError

from timetomodel import forecasting
from timetomodel.tests import utils as test_utils

"""
Test the forecasting functionality. As we need models and feature frames as well, these tests work well
as integration tests.
"""


DATA_START = datetime(2019, 1, 22, 15, tzinfo=pytz.UTC)
TOLERANCE = 0.01


def test_make_one_forecast():
    """ Given a simple linear model, try to make a forecast. """
    model, specs = test_utils.create_dummy_model_state(
        DATA_START, data_range_in_hours=24
    ).split()
    dt = datetime(2020, 1, 22, 22, tzinfo=pytz.timezone("Europe/Amsterdam"))
    # instead of using the 2019 data in our dummy model, we explicitly provide features ourselves
    # the model was trained on a linearly increasing range of integers, so
    # with lags 2 and 3 having values 892 and 891, we expect an outcome of 894
    # the model was trained on an external regressor that was always a constant 5, so
    # yet another 5 will provide no information
    features = pd.DataFrame(
        index=pd.date_range(start=dt, end=dt, freq="H"),
        data={"my_outcome_l2": 892, "my_outcome_l3": 891, "Regressor1": 5},
    )
    fc = forecasting.make_forecast_for(specs, features, model)
    assert abs(fc - 894) <= TOLERANCE


def test_make_one_forecast_with_transformation():
    """ Given a simple linear model with a transformation, try to make a forecast. """
    model, specs = test_utils.create_dummy_model_state(
        DATA_START,
        data_range_in_hours=24,
        outcome_feature_transformation=test_utils.MyAdditionTransformation(addition=7),
    ).split()
    dt = datetime(2020, 1, 22, 22, tzinfo=pytz.timezone("Europe/Amsterdam"))
    # instead of using the 2019 data in our dummy model, we explicitly provide features ourselves
    # the model was trained on a linearly increasing range of integers, so
    # with lags 2 and 3 having values 892 and 891, we expect an outcome of 894 (after the back-transformation)
    # the model was trained on an external regressor that was always a constant 5,
    # so yet another 5 will provide no information
    feature_data = specs.outcome_var.feature_transformation.transform_series(
        pd.Series([891, 892])
    )
    features = pd.DataFrame(
        index=pd.date_range(start=dt, end=dt, freq="H"),
        data={
            "my_outcome_l2": feature_data[1],
            "my_outcome_l3": feature_data[0],
            "Regressor1": 5,
        },
    )
    fc = forecasting.make_forecast_for(specs, features, model)
    assert abs(fc - 894) <= TOLERANCE


def test_rolling_forecast():
    """Using the simple linear model, create a rolling forecast"""
    model, specs = test_utils.create_dummy_model_state(
        DATA_START, data_range_in_hours=24
    ).split()
    h0 = 3  # first 3 hours can't be predicted,lacking the lagged outcome variable
    hn = 26  # only 2 additional forecast can be made, because the lowest lag is 2 hours
    start = DATA_START + timedelta(hours=h0)
    end = DATA_START + timedelta(hours=hn)
    forecasts = forecasting.make_rolling_forecasts(start, end, specs)[0]
    expected_values = range(h0, hn)
    for forecast, expected_value in zip(forecasts, expected_values):
        assert abs(forecast - expected_value) < TOLERANCE


def test_rolling_forecast_with_refitting(caplog):
    """Also rolling forecasting, but with re-fitting the model in between.
    We'll test if the expected number of re-fittings happened.
    Also, the model we end up with should not be the one we started with."""
    caplog.set_level(logging.DEBUG, logger="timetomodel.forecasting")
    model, specs = test_utils.create_dummy_model_state(
        DATA_START, data_range_in_hours=192
    ).split()
    start = DATA_START + timedelta(hours=70)
    end = DATA_START + timedelta(hours=190)
    forecasts, final_model_state = forecasting.make_rolling_forecasts(start, end, specs)
    expected_values = specs.outcome_var.load_series(
        time_window=(start, end),
        expected_frequency=timedelta(hours=1),
        check_time_window=True,
    ).loc[start:end][:-1]
    for forecast, expected_value in zip(forecasts, expected_values):
        assert abs(forecast - expected_value) < TOLERANCE
    refitting_logs = [
        log for log in caplog.records if "Fitting new model" in log.message
    ]
    remodel_frequency_in_hours = int(specs.remodel_frequency.total_seconds() / 3600)
    expected_log_times = [remodel_frequency_in_hours]
    while max(expected_log_times) < 190:
        expected_log_times.append(max(expected_log_times) + remodel_frequency_in_hours)
    assert len(refitting_logs) == len([elt for elt in expected_log_times if elt >= 70])
    assert model is not final_model_state.model


def test_missing_data_warning(caplog):

    caplog.set_level(logging.WARNING, logger="timetomodel.forecasting")
    with pytest.raises(MissingDataError):
        test_utils.create_dummy_model_state(
            DATA_START,
            data_range_in_hours=24,
            regressor_feature_transformation=test_utils.MyNanIntroducingTransformation(),
        )

    missing_data_logs = [log for log in caplog.records if "missing data" in log.message]
    assert len(missing_data_logs) == 1



from setuptools import setup

setup(
    name="timetomodel",
    description="Sane handling of time series data for forecast modelling - with production usage in mind.",
    author="Seita BV",
    author_email="nicolas@seita.nl",
    url="https://github.com/seitabv/timetomodel",
    keywords=["time series", "forecasting"],
    version="0.7.2",
    install_requires=[
        "pandas",
        "statsmodels",
        "scikit-learn",
        "matplotlib",
        "numpy",
        "scipy",
        "pytz",
        "python-dateutil >= 2.5",
        "SQLAlchemy",
    ],
    tests_require=["pytest"],
    packages=["timetomodel", "timetomodel.utils"],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    long_description="""\
# timetomodel

Time series forecasting is a modern data science & engineering challenge.

We noticed that these two worlds, data science and engineering of time series forecasting, are not very compatible.
Often, work from the data scientist has to be re-implemented by engineers to be used in production.

`timetomodel` was created to change that. It describes the data treatment of a model, and also automates common data treatment tasks like building data for training and testing.

As a *data scientist*, experiment with a model in your notebook.
Load data from static files (e.g. CSV) and try out lags, regressors and so on.
Compare plots and mean square errors of the models you developed.

As an *engineer*, take over the model description and use it in your production code.
Often, this would entail not much more than changing the data source (e.g from CSV to a column in the database).

`timetomodel` is supposed to wrap around any fit/predict type model, e.g. from statsmodels or scikit-learn (some work needed here to ensure support).


## Features

Here are some features for both data scientists and engineers to enjoy:

* Describe how to load data for outcome and regressor variables. Load from Pandas objects, CSV files, Pandas pickles or databases via SQLAlchemy.
* Create train & test data, including lags.
* Timezone awareness support.
* Custom data transformations, after loading (e.g. to remove duplicate) or only for forecasting (e.g. to apply a BoxCox transformation).
* Evaluate a model by RMSE, and plot the cumulative error.
* Support for creating rolling forecasts.


## Installation

``pip install timetomodel``

## Example

Here is an example where we describe a solar time series problem, and use ``statsmodels.OLS``, a linear regression model, to forecast one hour ahead:

    import pandas as pd
    import pytz
    from datetime import datetime, timedelta
    from statsmodels.api import OLS
    from timetomodel import speccing, ModelState, create_fitted_model, evaluate_models
    from timetomodel.transforming import BoxCoxTransformation
    from timetomodel.forecasting import make_rolling_forecasts

    data_start = datetime(2015, 3, 1, tzinfo=pytz.utc)
    data_end = datetime(2015, 10, 31, tzinfo=pytz.utc)

    #### Solar model - 1h ahead  ####

    # spec outcome variable
    solar_outcome_var_spec = speccing.CSVFileSeriesSpecs(
        file_path="data.csv",
        time_column="datetime",
        value_column="solar_power",
        name="solar power",
        feature_transformation=BoxCoxTransformation(lambda2=0.1)
    )
    # spec regressor variable
    regressor_spec_1h = speccing.CSVFileSeriesSpecs(
        file_path="data.csv",
        time_column="datetime",
        value_column="irradiation_forecast1h",
        name="irradiation forecast",
        feature_transformation=BoxCoxTransformation(lambda2=0.1)
    )
    # spec whole model treatment
    solar_model1h_specs = speccing.ModelSpecs(
        outcome_var=solar_outcome_var_spec,
        model=OLS,
        frequency=timedelta(minutes=15),
        horizon=timedelta(hours=1),
        lags=[lag * 96 for lag in range(1, 8)],  # 7 days (data has daily seasonality)
        regressors=[regressor_spec_1h],
        start_of_training=data_start + timedelta(days=30),
        end_of_testing=data_end,
        ratio_training_testing_data=2/3,
        remodel_frequency=timedelta(days=14)  # re-train model every two weeks
    )

    solar_model1h = create_fitted_model(solar_model1h_specs, "Linear Regression Solar Horizon 1h")
    # solar_model_1h is now an OLS model object which can be pickled and re-used.
    # With the solar_model1h_specs in hand, your production code could always re-train a new one,
    # if the model has become outdated.

    # For data scientists: evaluate model
    evaluate_models(m1=ModelState(solar_model1h, solar_model1h_specs))

![Evaluation result](https://raw.githubusercontent.com/SeitaBV/timetomodel/master/img/solar-forecast-evaluation.png)

    # For engineers a): Change data sources to use database (hinted)
    solar_model1h_specs.outcome_var = speccing.DBSeriesSpecs(query=...)
    solar_model1h_specs.regressors[0] = speccing.DBSeriesSpecs(query=...)

    # For engineers b): Use model to make forecasts for an hour
    forecasts, model_state = make_rolling_forecasts(
        start=datetime(2015, 11, 1, tzinfo=pytz.utc),
        end=datetime(2015, 11, 1, 1, tzinfo=pytz.utc),
        model_specs=solar_model1h_specs
    )
    # model_state might have re-trained a new model automatically, by honoring the remodel_frequency

    """,
    long_description_content_type="text/markdown",
)



from datetime import datetime, timedelta

import pandas as pd
import pytest
import pytz

from timetomodel.utils.time_utils import (
    get_closest_quarter,
    get_most_recent_quarter,
    round_datetime,
    timedelta_fits_into,
    timedelta_to_pandas_freq_str,
)


def test_find_quarter():
    dt = datetime(2018, 1, 26, 14, 40).astimezone(pytz.timezone("Europe/Amsterdam"))
    recent = get_most_recent_quarter(dt)
    assert recent.day == dt.day
    assert recent.hour == dt.hour
    assert recent.minute == 30
    closest = get_closest_quarter(dt)
    assert closest.day == dt.day
    assert closest.hour == dt.hour
    assert closest.minute == 45


def test_timedelta_to_pd_freq_str():
    assert timedelta_to_pandas_freq_str(timedelta(seconds=5)) == "5S"
    assert timedelta_to_pandas_freq_str(timedelta(minutes=2)) == "2T"
    assert timedelta_to_pandas_freq_str(timedelta(minutes=15)) == "15T"
    assert timedelta_to_pandas_freq_str(timedelta(hours=1)) == "H"
    assert timedelta_to_pandas_freq_str(timedelta(hours=2)) == "2H"
    assert timedelta_to_pandas_freq_str(timedelta(hours=26)) == "26H"
    assert timedelta_to_pandas_freq_str(timedelta(days=1, hours=2)) == "26H"
    assert timedelta_to_pandas_freq_str(timedelta(days=1)) == "D"


@pytest.mark.parametrize(
    "dt",
    [
        datetime(2018, 1, 26, 14, 40),
        datetime(2018, 1, 26, 14, 40, tzinfo=pytz.utc),
        pd.Timestamp(datetime(2018, 1, 26, 14, 40)),
        pd.Timestamp(datetime(2018, 1, 26, 14, 40, tzinfo=pytz.utc)),
    ],
)
def test_round_time_by_hour(dt):
    round_to_hour = round_datetime(dt, by_seconds=60 * 60)
    assert round_to_hour.day == dt.day
    assert round_to_hour.hour == 15
    assert round_to_hour.minute == 00


@pytest.mark.parametrize(
    "dt",
    [
        datetime(2018, 1, 26, 14, 40),
        datetime(2018, 1, 26, 14, 40, tzinfo=pytz.utc),
        pd.Timestamp(datetime(2018, 1, 26, 14, 40)),
        pd.Timestamp(datetime(2018, 1, 26, 14, 40, tzinfo=pytz.utc)),
    ],
)
def test_round_time_by_15min(dt):
    round_to_hour = round_datetime(dt, by_seconds=60 * 15)
    assert round_to_hour.day == dt.day
    assert round_to_hour.hour == 14
    assert round_to_hour.minute == 45


def test_timedelta_fits():
    assert not timedelta_fits_into(timedelta(seconds=11), timedelta(minutes=4))
    assert timedelta_fits_into(timedelta(minutes=10), timedelta(hours=2))
    assert timedelta_fits_into(timedelta(minutes=3), timedelta(hours=1))
    assert timedelta_fits_into(timedelta(minutes=15), timedelta(hours=1))
    assert timedelta_fits_into(timedelta(minutes=15), timedelta(days=4))
    assert timedelta_fits_into(timedelta(hours=12), timedelta(days=2))
    assert not timedelta_fits_into(timedelta(hours=16), timedelta(days=3))
    assert timedelta_fits_into(timedelta(hours=16), timedelta(days=6))
    assert timedelta_fits_into(timedelta(minutes=15), timedelta(weeks=1))
    assert not timedelta_fits_into(timedelta(minutes=11), timedelta(hours=1))
    assert timedelta_fits_into(timedelta(minutes=11), timedelta(hours=11))



from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import pytest
import pytz

from timetomodel.exceptions import IncompatibleModelSpecs, MissingData, NaNData
from timetomodel.speccing import CSVFileSeriesSpecs, ObjectSeriesSpecs
from timetomodel.tests.utils import MyMultiplicationTransformation
from timetomodel.transforming import Transformation


def test_load_series_without_datetime_index():
    with pytest.raises(Exception) as e_info:
        s = ObjectSeriesSpecs(data=pd.Series([1, 2, 3]), name="mydata")
        s.load_series(expected_frequency=timedelta(hours=1))
    assert "DatetimeIndex" in str(e_info.value)


def test_load_series():
    dt = datetime(2019, 1, 29, 15, 15)
    s = ObjectSeriesSpecs(
        data=pd.Series(
            index=pd.date_range(dt, dt + timedelta(minutes=30), freq="15T"),
            data=[1, 2, 3],
        ),
        name="mydata",
    )
    assert (
        s.load_series(
            time_window=(dt, dt + timedelta(minutes=30)),
            expected_frequency=timedelta(minutes=15),
            check_time_window=True,
        ).loc[dt + timedelta(minutes=30)]
        == 3
    )


def test_load_series_with_expected_time_window():
    dt = datetime(2019, 1, 29, 15, 15, tzinfo=pytz.utc)
    s = ObjectSeriesSpecs(
        data=pd.Series(
            index=pd.date_range(dt, dt + timedelta(minutes=30), freq="15T"),
            data=[1, 2, 3],
        ),
        name="mydata",
    )
    assert (
        s.load_series(
            time_window=(dt, dt + timedelta(minutes=30)),
            expected_frequency=timedelta(minutes=15),
            check_time_window=True,
        ).loc[dt + timedelta(minutes=30)]
        == 3
    )


def test_load_series_with_larger_expected_time_window():
    dt = datetime(2019, 1, 29, 15, 15, tzinfo=pytz.utc)
    s = ObjectSeriesSpecs(
        data=pd.Series(
            index=pd.date_range(dt, dt + timedelta(minutes=30), freq="15T"),
            data=[1, 2, 3],
        ),
        name="mydata",
    )
    with pytest.raises(MissingData) as e_info:
        s.load_series(
            expected_frequency=timedelta(minutes=15),
            time_window=(
                dt - timedelta(minutes=15),
                dt + timedelta(minutes=45),
            ),
            check_time_window=True,
        )
    assert "starts too late" in str(e_info.value)
    assert "ends too early" in str(e_info.value)


@pytest.mark.parametrize("down_or_up", ["down", "up"])
def test_load_series_with_frequency_resampling(down_or_up: str):
    dt = datetime(2019, 1, 29, 15, 15)
    s = ObjectSeriesSpecs(
        data=pd.Series(
            index=pd.date_range(dt, dt + timedelta(minutes=30), freq="15T"),
            data=[1, 2, 3],
        ),
        name="mydata",
    )
    series = s.load_series(
        expected_frequency=timedelta(hours=1)
        if down_or_up == "down"
        else timedelta(minutes=5)
    )
    assert len(series) == 1 if down_or_up == "down" else len(series) == 9
    assert series.mean() == 2  # the mean remains the same


@pytest.mark.parametrize("down_or_up", ["down", "up"])
def test_load_series_with_non_existing_custom_frequency_resampling(down_or_up: str):
    dt = datetime(2019, 1, 29, 15, 15)
    s = ObjectSeriesSpecs(
        data=pd.Series(
            index=pd.date_range(dt, dt + timedelta(minutes=30), freq="15T"),
            data=[1, 2, 3],
        ),
        name="mydata",
        resampling_config={f"{down_or_up}sampling_method": "GGG"},
    )

    with pytest.raises(IncompatibleModelSpecs) as e_info:
        s.load_series(
            expected_frequency=timedelta(hours=1)
            if down_or_up == "down"
            else timedelta(minutes=5)
        )
    assert f"Cannot find {down_or_up}sampling method GGG" in str(e_info.value)


@pytest.mark.parametrize("down_or_up", ["down", "up"])
def test_load_series_with_custom_frequency_resampling(down_or_up: str):
    dt = datetime(2019, 1, 29, 15, 15)
    s = ObjectSeriesSpecs(
        data=pd.Series(
            index=pd.date_range(dt, dt + timedelta(minutes=30), freq="15T"),
            data=[1, 2, 3],
        ),
        name="mydata",
        resampling_config={
            "downsampling_method": "sum",
            "upsampling_method": "reverse_sum",
        },
    )

    series = s.load_series(
        expected_frequency=timedelta(hours=1)
        if down_or_up == "down"
        else timedelta(minutes=5)
    )
    assert len(series) == 1 if down_or_up == "down" else len(series) == 9
    assert sum(series) == 6  # the sum remains the same


@pytest.mark.parametrize(
    "down_or_up, exp_series",
    [
        (
            "down",
            pd.Series(
                [1, 3],
                index=pd.date_range(
                    datetime(2019, 1, 29, 15),
                    datetime(2019, 1, 29, 16),
                    freq="1H",
                    tz="UTC",
                ),
            ),
        ),
        (
            "up",
            pd.Series(
                [1, 1, 1, 2, 2, 2, 3],
                index=pd.date_range(
                    datetime(2019, 1, 29, 15, 30),
                    datetime(2019, 1, 29, 16),
                    freq="5T",
                    tz="UTC",
                ),
                dtype="float64",
            ),
        ),
    ],
)
def test_load_series_with_instantaneous_measurements(down_or_up: str, exp_series):
    """ Test resampling of instantaneous measurements. """
    dt = datetime(2019, 1, 29, 15, 30)
    s = ObjectSeriesSpecs(
        data=pd.Series(
            index=pd.date_range(dt, dt + timedelta(minutes=30), freq="15T"),
            data=[1, 2, 3],
        ),
        name="mydata",
        resampling_config=dict(
            downsampling_method="first",
            event_resolution=timedelta(hours=0),
        ),
        interpolation_config=dict(method="pad", limit=3),
    )

    series = s.load_series(
        expected_frequency=timedelta(hours=1)
        if down_or_up == "down"
        else timedelta(minutes=5)
    )
    assert len(series) == 2 if down_or_up == "down" else len(series) == 7
    pd.testing.assert_series_equal(series, exp_series)


def test_load_series_without_data():
    dt = datetime(2019, 1, 29, 15, 15)
    s = ObjectSeriesSpecs(
        data=pd.Series(
            index=pd.date_range(dt, dt + timedelta(minutes=30), freq="15T"),
            data=[np.nan, np.nan, np.nan],
        ),
        name="mydata",
    )
    with pytest.raises(NaNData) as e_info:
        s.load_series(expected_frequency=timedelta(hours=1))
    assert "Nan values" in str(e_info.value)


def test_load_series_with_missing_data():
    dt = datetime(2019, 1, 29, 15, 15)
    s = ObjectSeriesSpecs(
        data=pd.Series(
            index=pd.date_range(dt, dt + timedelta(minutes=30), freq="15T"),
            data=[1, np.nan, 3],
        ),
        name="mydata",
    )
    with pytest.raises(NaNData) as e_info:
        s.load_series(expected_frequency=timedelta(minutes=15))
    assert "Nan values" in str(e_info.value)


def test_load_series_with_transformation():
    dt = datetime(2019, 1, 29, 15, 15)
    s = ObjectSeriesSpecs(
        data=pd.Series(
            index=pd.date_range(dt, dt + timedelta(minutes=30), freq="15T"),
            data=[1, 2, 3],
        ),
        name="mydata",
        feature_transformation=MyMultiplicationTransformation(factor=11),
    )
    assert (
        s.load_series(expected_frequency=timedelta(minutes=15)).loc[
            dt + timedelta(minutes=15)
        ]
        == 2
    )
    assert (
        s.load_series(
            expected_frequency=timedelta(minutes=15), transform_features=True
        ).loc[dt + timedelta(minutes=15)]
        == 2 * 11
    )


def test_load_series_from_csv_with_post_load_processing(tmpdir):

    highscore_data = """Time,Name,Highscore,
2019-02-05T12:57:00,Mel,8,
2019-02-05T10:30:00,Jack,5,
2019-02-05T11:36:00,David,10,
2019-02-05T10:34:00,Peter,6,
2019-02-05T09:11:00,David,5,
2019-02-05T11:17:00,Ryan,9,
2019-02-05T12:27:00,Ryan,9,
"""
    f = tmpdir.join("highscore.csv")
    f.write(highscore_data)

    def to_hour(dt: datetime) -> datetime:
        return dt.replace(minute=0, second=0, microsecond=0)

    class BestHighscorePerHour(Transformation):
        def transform_dataframe(self, df):
            df["Time"] = pd.to_datetime(df["Time"], utc=True)
            df["Time"] = df["Time"].apply(to_hour)

            return (
                df.sort_values(by=["Highscore"], ascending=False)
                .drop_duplicates(subset=["Time"], keep="first")
                .sort_values(by=["Time"])
            )

    s = CSVFileSeriesSpecs(
        file_path=f.realpath(),
        time_column="Time",
        value_column="Highscore",
        post_load_processing=BestHighscorePerHour(),
        name="mydata",
        feature_transformation=MyMultiplicationTransformation(factor=100),
    )

    data = s.load_series(expected_frequency=timedelta(hours=1))

    assert data[datetime(2019, 2, 5, 9)] == 5
    assert data[datetime(2019, 2, 5, 10)] == 6
    assert data[datetime(2019, 2, 5, 11)] == 10
    assert data[datetime(2019, 2, 5, 12)] == 9

    data = s.load_series(expected_frequency=timedelta(hours=1), transform_features=True)

    assert data[datetime(2019, 2, 5, 9)] == 500
    assert data[datetime(2019, 2, 5, 10)] == 600
    assert data[datetime(2019, 2, 5, 11)] == 1000
    assert data[datetime(2019, 2, 5, 12)] == 900


def test_load_series_with_non_existing_interpolation():
    dt = datetime(2019, 1, 29, 15, 15)
    s = ObjectSeriesSpecs(
        data=pd.Series(
            index=pd.date_range(dt, dt + timedelta(minutes=30), freq="15T"),
            data=[1, np.nan, 3],
        ),
        name="mydata",
        interpolation_config={"method": "GGG"},
    )

    with pytest.raises(IncompatibleModelSpecs) as e_info:
        s.load_series(expected_frequency=timedelta(minutes=15))
    assert "Cannot call interpolate function with arguments {'method': 'GGG'}" in str(
        e_info.value
    )


def test_load_series_with_interpolation():
    dt = datetime(2019, 1, 29, 15, 15)
    s = ObjectSeriesSpecs(
        data=pd.Series(
            index=pd.date_range(dt, dt + timedelta(minutes=30), freq="15T"),
            data=[1, np.nan, 3],
        ),
        name="mydata",
        interpolation_config={"method": "time"},
    )

    series = s.load_series(expected_frequency=timedelta(minutes=15))
    assert len(series) == 3
    assert series[1] == 2  # the interpolated value
